@@ -122,6 +122,7 @@ static int network_ssl_servername_callback(SSL *ssl, int *al, server *srv) {
 	config_setup_connection(srv, con);
 
 	con->conditional_is_valid[COMP_SERVER_SOCKET] = 1;
+	con->conditional_is_valid[COMP_HTTP_REMOTE_IP] = 1;
 	con->conditional_is_valid[COMP_HTTP_SCHEME] = 1;
 	con->conditional_is_valid[COMP_HTTP_HOST] = 1;
 	config_patch_connection(srv, con);
@@ -769,9 +770,18 @@ int network_init(server *srv) {
 		if (buffer_string_is_empty(s->ssl_pemfile) && buffer_string_is_empty(s->ssl_ca_file)) continue;
 
 		if (srv->ssl_is_init == 0) {
+		      #if OPENSSL_VERSION_NUMBER >= 0x10100000L \
+		       && !defined(LIBRESSL_VERSION_NUMBER)
+			OPENSSL_init_ssl(OPENSSL_INIT_LOAD_SSL_STRINGS
+					|OPENSSL_INIT_LOAD_CRYPTO_STRINGS,NULL);
+			OPENSSL_init_crypto(OPENSSL_INIT_ADD_ALL_CIPHERS
+					   |OPENSSL_INIT_ADD_ALL_DIGESTS
+					   |OPENSSL_INIT_LOAD_CONFIG, NULL);
+		      #else
 			SSL_load_error_strings();
 			SSL_library_init();
 			OpenSSL_add_all_algorithms();
+		      #endif
 			srv->ssl_is_init = 1;
 
 			if (0 == RAND_status()) {
@@ -831,7 +841,7 @@ int network_init(server *srv) {
 		SSL_CTX_set_options(s->ssl_ctx, ssloptions);
 		SSL_CTX_set_info_callback(s->ssl_ctx, ssl_info_callback);
 
-		if (!s->ssl_use_sslv2) {
+		if (!s->ssl_use_sslv2 && 0 != SSL_OP_NO_SSLv2) {
 			/* disable SSLv2 */
 			if ((SSL_OP_NO_SSLv2 & SSL_CTX_set_options(s->ssl_ctx, SSL_OP_NO_SSLv2)) != SSL_OP_NO_SSLv2) {
 				log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:",
@@ -840,7 +850,7 @@ int network_init(server *srv) {
 			}
 		}
 
-		if (!s->ssl_use_sslv3) {
+		if (!s->ssl_use_sslv3 && 0 != SSL_OP_NO_SSLv3) {
 			/* disable SSLv3 */
 			if ((SSL_OP_NO_SSLv3 & SSL_CTX_set_options(s->ssl_ctx, SSL_OP_NO_SSLv3)) != SSL_OP_NO_SSLv3) {
 				log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:",
